"""
Accessibility audit feature
"""

from .a11y_audit import A11yAudit, AccessibilityError
from .axe_core_ruleset import AxeCoreAudit, AxeCoreAuditConfig
from .axs_ruleset import AxsAudit, AxsAuditConfig



"""
Base class for testing a web application.
"""
from abc import ABCMeta
import functools
import os
import sys
from unittest import SkipTest
from uuid import uuid4

from needle.cases import NeedleTestCase, import_from_string
from selenium.webdriver.support.events import EventFiringWebDriver

from .browser import browser, save_screenshot, save_driver_logs
from .proxy import bmp_proxy, stop_server
from .performance import HarListener, HarCapturer


class WebAppTest(NeedleTestCase):

    """
    Base class for testing a web application.
    """

    __metaclass__ = ABCMeta

    # Execute tests in parallel!
    _multiprocess_can_split_ = True

    def __init__(self, *args, **kwargs):
        super(WebAppTest, self).__init__(*args, **kwargs)

        # This allows using the @attr() decorator from nose to set these on a
        # test by test basis
        self.proxy = getattr(self, 'proxy', None)
        self.har_mode = getattr(
            self, 'har_mode', os.environ.get('BOK_CHOY_HAR_MODE', None))

    @classmethod
    def setUpClass(cls):
        """
        Override NeedleTestCase's setUpClass method so that it does not
        start up the browser once for each testcase class.
        Instead we start up the browser once per TestCase instance,
        in the setUp method.
        """
        # Instantiate the diff engine.
        # This will allow Needle's flexibility for choosing which you want to use.
        # These lines are copied over from Needle's setUpClass method.
        klass = import_from_string(cls.engine_class)
        cls.engine = klass()

        # Needle's setUpClass method set up the driver (thus starting up the browser),
        # and set the initial window position and viewport size.
        # Those lines are not copied here into WebAppTest's setUpClass method,
        # but instead into our setUp method. This follows our paradigm of starting
        # up a new browser session for each TestCase.

        # Now call the super of the NeedleTestCase class, so that we get everything
        # from the setUpClass method of its parent (unittest.TestCase).
        super(NeedleTestCase, cls).setUpClass()  # pylint: disable=bad-super-call

    @classmethod
    def tearDownClass(cls):
        """
        Override NeedleTestCase's tearDownClass method because it
        would quit the browser. This is not needed as we have already quit the browser
        after each TestCase, by virtue of a cleanup that we add in the setUp method.
        """
        # We still want to call the super of the NeedleTestCase class, so that we get
        # everything from the tearDownClass method of its parent (unittest.TestCase).
        super(NeedleTestCase, cls).tearDownClass()  # pylint: disable=bad-super-call

    def get_web_driver(self):
        """
        Override NeedleTestCases's get_web_driver class method to return the WebDriver instance
        that is already being used, instead of starting up a new one.
        """
        return self.browser

    def set_viewport_size(self, width, height):
        """
        Override NeedleTestCases's set_viewport_size class method because we need it to operate
        on the instance not the class.

        See the Needle documentation at http://needle.readthedocs.org/ for information on this
        feature. It is particularly useful to predict the size of the resulting screenshots
        when taking fullscreen captures, or to test responsive sites.
        """
        self.driver.set_window_size(width, height)

        # Measure the difference between the actual document width and the
        # desired viewport width so we can account for scrollbars:
        script = "return {width: document.body.clientWidth, height: document.body.clientHeight};"
        measured = self.driver.execute_script(script)
        delta = width - measured['width']

        self.driver.set_window_size(width + delta, height)

    def setUp(self):
        """
        Start the browser for use by the test.
        You *must* call this in the `setUp` method of any subclasses before using the browser!

        Returns:
            None
        """
        super(WebAppTest, self).setUp()

        if self.har_mode:
            # Set up proxy using browsermobproxy if we want to capture har files
            # or if the user has specified that it wants to use browsermobproxy
            self.proxy, server = bmp_proxy()
            self.addCleanup(stop_server, server)

        # Set up the browser
        # This will start the browser
        # If using SauceLabs, tag the job with test info
        tags = [self.id()]
        self.browser = browser(tags, self.proxy)

        # Needle uses these attributes for taking the screenshots
        self.driver = self.get_web_driver()
        self.driver.set_window_position(0, 0)
        self.set_viewport_size(self.viewport_width, self.viewport_height)

        if self.har_mode:
            # Initialize a HarCapturer. The har_capturer instance will always be
            # able to be explicitly interacted with in the test in addition to
            # being accessible to the HarListener for automatic capture.
            self.har_capturer = HarCapturer(
                self.proxy,
                har_base_name=self.id(),
                mode=self.har_mode,
            )

            # In order to automatically capture the har, we need a listener that can
            # track the pages visted. To do this, we can use an
            # EventFiringWebdriver.
            self.browser = EventFiringWebDriver(
                self.browser, HarListener(self.har_capturer))

        # Cleanups are executed in LIFO order.
        # This ensures that the screenshot is taken and the driver logs are saved
        # BEFORE the browser quits.
        self.addCleanup(self.browser.quit)
        self.addCleanup(self._save_artifacts)

    @property
    def unique_id(self):
        """
        Helper method to return a uuid.

        Returns:
            39-char UUID string
        """
        return str(uuid4().int)  # pylint: disable=no-member

    def _save_artifacts(self):
        """
        On failure or error save a screenshot, the
        selenium driver logs, and the captured har file.
        """
        # Determine whether the test case succeeded or failed
        result = sys.exc_info()
        exception_type = result[0]

        # Do not save artifacts for skipped tests.
        if exception_type is SkipTest:
            return

        # If it failed, take a screenshot and save the driver logs.
        # The exception info will either be an assertion error (on failure)
        # or an actual exception (on error)
        if result != (None, None, None):
            try:
                save_screenshot(self.browser, self.id())
            except:  # pylint: disable=bare-except
                pass

            try:
                save_driver_logs(self.browser, self.id())
            except:  # pylint: disable=bare-except
                pass

            try:
                self.har_capturer.save_har(
                    self.browser,
                    caller_mode=self.har_mode
                )
            except:  # pylint: disable=bare-except
                pass


def with_cache(function):
    """
    A decorator to be used on a test case of a WebAppTest test to run the test twice in
    the same browser instance, capturing a har file each time.

    The first har captured will reflect the performance when there have not been any
    assets cached. The second will reflect the performance of a second visit, having
    possibly cached some assets in the first visit.

    Note that this will only work in 'explicit' and 'auto' modes. Using with `BOK_CHOY_HAR_MODE`
    set to 'error' will only save if the test fails or errors.

    Args:
        function (callable): The function to decorate. It should be a test_case ina a
        WebAppTest instance.

    Returns:
        Decorated method
    """

    @functools.wraps(function)
    def wrapper(self, *args, **kwargs):
        """
        Runs the test case twice. The first time, there will be an empty cache. The second
        time, the cache will contain anything stored on the first call.
        """
        # Run once in a new browser instance.
        function(self, *args, **kwargs)

        self.har_capturer.save_har(self.browser, caller_mode=self.har_mode)
        self.har_capturer._with_cache = True  # pylint: disable=protected-access
        # Run the whole thing again in the same browser instance.
        function(self, *args, **kwargs)

    return wrapper



"""
Utilities for checking network performance of a web application.
"""
import os
import json
import datetime
from textwrap import dedent
from selenium.webdriver.support.events import AbstractEventListener

HAR_CAPTURE_MODES = ('error', 'explicit', 'auto')


class MethodNotEnabledInCurrentMode(Exception):
    """
    An exception for when a method is called that should not be used
    in the current mode.
    """
    def __str__(self):
        return (
            "To manipulate the har on your own, please use @attr(har_mode='"
            "explicit') or BOK_CHOY_HAR_MODE='explicit'."
        )

class UnknownHarCaptureMode(Exception):
    """
    An exception for when the mode selected doesn't match any defined
    modes.
    """
    def __str__(self):
        return 'Expected one of: {}'.format(HAR_CAPTURE_MODES)


class HarListener(AbstractEventListener):
    """
    An object that can automatically track and save a HAR file from the
    selenium controlled browser.

    Usage:
        driver = EventFiringWebDriver(driver, HarListener(har_capturer))

    Args:
        har_capturer: An instance of HarCapturer.

    """
    def __init__(self, har_capturer, *args, **kwargs):  # pylint: disable=super-on-old-class
        super(HarListener, self).__init__(*args, **kwargs)
        self.har_capturer = har_capturer

    def before_navigate_to(self, url, driver):  # pylint: disable=missing-docstring
        if self.har_capturer.mode in ('auto', 'error'):
            self.har_capturer.add_page(driver, url, caller_mode=self.har_capturer.mode)

    def before_close(self, driver):  # pylint: disable=missing-docstring
        if self.har_capturer.mode in ('auto',):
            self.har_capturer.save_har(driver, caller_mode=self.har_capturer.mode)

    def before_quit(self, driver):  # pylint: disable=missing-docstring
        if self.har_capturer.mode in ('auto',):
            self.har_capturer.save_har(driver, caller_mode=self.har_capturer.mode)


class HarCapturer(object):
    """
    An object that can track and manipulate HAR files.

    Args:
        proxy: a browsermobproxy proxy instance.

    Keyword Args:
        har_base_name: A base for the har file names. The har filename may
            still have the datetime and 'cached' specifier appended to them.

        mode: 'auto', 'error', or 'explicit'. Although it is a kwarg, an
            UnknownHarCaptureMode exception will be raised if no mode is passed
            or if the mode doesn't match one of the following.

            * auto: automatically save a single har file for each test.
            * error: automatically save a single har file for each test only if
                it fails or errors.
            * explicit: interact explicitly with this object in order to capture
                and save a har file.
    """
    def __init__(self, *args, **kwargs):
        super(HarCapturer, self).__init__()

        # We need to access the proxy server in addition to the browser.
        self.proxy = args[0]
        self._har_base_name = kwargs.get('har_base_name', '')

        self.mode = kwargs.get('mode', None)

        if self.mode not in HAR_CAPTURE_MODES:
            raise UnknownHarCaptureMode

        # Vars for tracking state
        self._page_timings = []
        self._active_har = False
        self._with_cache = False

    def _validate_mode(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Raises MethodNotEnabledInCurrentMode if a user tries to interact
        with the har explicitly while in 'auto' or 'error' mode.
        """
        caller_mode = kwargs.get('caller_mode', 'explicit')
        if caller_mode != self.mode:
            raise MethodNotEnabledInCurrentMode

    def add_page(self, driver, page_name, *args, **kwargs):
        """
        Creates a new page within the har file. If no har file has been started
        since the last save, it will create one.

        Args:
            page_name: a string to be used as the page name in the har

        Returns:
            None
        """
        self._validate_mode(*args, **kwargs)

        if not self._active_har:
            # Start up a new HAR file
            self.proxy.new_har(
                ref=page_name,
                options={
                    'captureContent': True,
                    'captureHeaders': True,
                    'captureBinaryContent': True,
                }
            )

            self._page_timings = []
            self._active_har = True
        else:
            # Save the timings for the previous page before moving on to recording the new one.
            self._record_page_timings(driver)
            self.proxy.new_page(ref=page_name)

    def _record_page_timings(self, driver):
        """
        Saves recorded page timings to self.timings to be added to the har file
        on save.

        Returns:
            None
        """

        script = dedent("""
            var performance = window.performance || {};
            var timings = performance.timing || {};
            return timings;
        """)

        # Capture the timings from the browser via javascript
        timings = driver.execute_script(script)
        self._page_timings.append(timings)


    def har_name(self, name_override=None):
        """
        Returns the name to use for a saved artifacts.
        """

        if name_override:
            file_name = name_override

        else:
            file_name = "{}_{}".format(
                self._har_base_name,
                datetime.datetime.utcnow().isoformat()
            )

        if self._with_cache:
            file_name += '_cached'

        return file_name

    def save_har(self, driver, name_override=None, *args, **kwargs):
        """
        Save a HAR file.

        The location of the har file can be configured
        by the environment variable `BOK_CHOY_HAR_DIR`.  If not set,
        this defaults to the current working directory.

        Returns:
            None
        """
        self._validate_mode(*args, **kwargs)

        if self._active_har:
            # Record the most recent pages timings
            self._record_page_timings(driver)
            timings = self._page_timings

            # Get the har contents from the proxy
            har = self.proxy.har

            # Record the timings from the pages
            for index, _timing in enumerate(timings):
                nav_start = timings[index]['navigationStart']
                dom_content_loaded_event_end = timings[index]['domContentLoadedEventEnd']
                load_event_end = timings[index]['loadEventEnd']
                har['log']['pages'][index]['pageTimings']['onContentLoad'] = dom_content_loaded_event_end - nav_start
                har['log']['pages'][index]['pageTimings']['onLoad'] = load_event_end - nav_start

            har_file = os.path.join(
                os.environ.get('BOK_CHOY_HAR_DIR', ''),
                '{}.har'.format(self.har_name(name_override)))
            with open(har_file, 'w') as output_file:
                json.dump(har, output_file)
                output_file.close()

            # Set this to false so that a new har will be started if new_page is called again
            self._active_har = False



"""
Interface for using the axe-core ruleset.
See: https://github.com/dequelabs/axe-core
"""
import json
import os
import requests

from textwrap import dedent, fill

from .a11y_audit import A11yAudit, A11yAuditConfig, AccessibilityError
from ..promise import Promise


CUR_DIR = os.path.dirname(os.path.abspath(__file__))


class AxeCoreAuditConfig(A11yAuditConfig):
    """
    The `AxeCoreAuditConfig` object defines the options available when
    running an `AxeCoreAudit`.
    """
    def __init__(self, *args, **kwargs):
        super(AxeCoreAuditConfig, self).__init__(*args, **kwargs)
        self.rules, self.context = None, None
        self.rules_file = os.path.join(
            os.path.split(CUR_DIR)[0],
            'vendor/axe-core/axe.min.js'
        )
        self.set_rules({})
        self.set_scope()

    def set_rules(self, rules):
        """
        Set rules to ignore XOR limit to when checking for accessibility
        errors on the page.

        Args:

            rules: a dictionary one of the following formats.
                If you want to run all of the rules except for some::

                    {"ignore": []}

                If you want to run only a specific set of rules::

                    {"apply": []}

                If you want to run only rules of a specific standard::

                    {"tags": []}

        Examples:

            To run only "bad-link" and "color-contrast" rules::

                page.a11y_audit.config.set_rules({
                    "apply": ["bad-link", "color-contrast"],
                })

            To run all rules except for "bad-link" and "color-contrast"::

                page.a11y_audit.config.set_rules({
                    "ignore": ["bad-link", "color-contrast"],
                })

            To run only WCAG 2.0 Level A rules::

                page.a11y_audit.config.set_rules({
                    "tags": ["wcag2a"],
                })

            To run all rules:
                page.a11y_audit.config.set_rules({})

        Related documentation:

            * https://github.com/dequelabs/axe-core/blob/master/doc/API.md#options-parameter-examples
            * https://github.com/dequelabs/axe-core/doc/rule-descriptions.md
        """
        options = {}
        if rules:
            if rules.get("ignore"):
                options["rules"] = {}
                for rule in rules.get("ignore"):
                    options["rules"][rule] = {"enabled": False}
            elif rules.get("apply"):
                options["runOnly"] = {
                    "type": "rule",
                    "values": rules.get("apply"),
                }
            elif rules.get("tags"):
                options["runOnly"] = {
                    "type": "tag",
                    "values": rules.get("tags"),
                }
        self.rules = json.dumps(options)

    def set_scope(self, include=None, exclude=None):
        """
        Sets `scope` (refered to as `context` in ruleset documentation), which
        defines the elements on a page to include or exclude in the audit. If
        neither `include` nor `exclude` are passed, the entire document will
        be included.

        Args:

            include (optional): a list of css selectors for elements that
            should be included in the audit. By, default, the entire document
            is included.
            exclude (optional): a list of css selectors for elements that should not
            be included in the audit.

        Examples:

            To include all items in `#main-content` except `#some-special-elm`::

                page.a11y_audit.config.set_scope(
                    exclude=["#some-special-elm"],
                    include=["#main-content"]
                )

            To include all items in the document except `#some-special-elm`::

                page.a11y_audit.config.set_scope(
                    exclude=["#some-special-elm"],
                )

            To include only children of `#some-special-elm`::

                page.a11y_audit.config.set_scope(
                    include=["#some-special-elm"],
                )

        Context documentation:

            https://github.com/dequelabs/axe-core/blob/master/doc/API.md#a-context-parameter

            Note that this implementation only supports css selectors. It does
            not accept nodes as described in the above documentation resource.
        """
        context = {}

        if exclude:
            context["exclude"] = [[selector] for selector in exclude]

        if include:
            context["include"] = [[selector] for selector in include]

        self.context = json.dumps(context) if context else 'document'


class AxeCoreAudit(A11yAudit):
    """
    Use Deque Labs' axe-core engine to audit a page for accessibility issues.

    Since this needs to inject JavaScript into the browser page, the only
    known way to do this is to use PhantomJS as your browser.

    Related documentation:

        https://github.com/dequelabs/axe-core/blob/master/doc/API.md
    """

    @property
    def default_config(self):
        """
        Returns an instance of a subclass of AxeCoreAuditConfig.
        """
        return AxeCoreAuditConfig()

    @staticmethod
    def _check_rules(ghostdriver_url, session_id, config):
        """
        Run an accessibility audit on the page using the axe-core ruleset.

        Args:
            ghostdriver_url: url of ghostdriver.
            session_id: a session id to test.
            config: an AxsAuditConfig instance.

        Returns:
            A list of violations.

        Related documentation:

            https://github.com/dequelabs/axe-core/blob/master/doc/API.md#results-object

        __Caution__: You probably don't really want to call this method
        directly! It will be used by `AxeCoreAudit.do_audit`.
        """
        script_url = '{}/session/{}/phantom/execute'.format(
            ghostdriver_url, session_id)

        audit_run_script = dedent("""
            return this.evaluate(function(){{
                var updatedResults = function(r) {{
                    window.a11yAuditResults = JSON.stringify(r);
                    window.console.log(window.a11yAuditResults);
                }}

                axe.a11yCheck({context}, {options}, updatedResults);
            }});
        """.format(context=config.context, options=config.rules))

        audit_results_script = dedent("""
            return this.evaluate(function(){{
                window.console.log(window.a11yAuditResults);
                return window.a11yAuditResults;
            }});
        """)

        requests.post(
            script_url,
            data=json.dumps({"script": audit_run_script, "args": []}),
        )

        def audit_results_check_func():
            """
            A method to check that the audit has completed.

            Returns:

                (True, results) if the results are available.
                (False, None) if the results aren't available.
            """
            resp = requests.post(
                script_url,
                data=json.dumps({"script": audit_results_script, "args": []})
            )

            try:
                results = json.loads(resp.json().get('value'))
            except (TypeError, ValueError):
                results = None

            if results:
                return True, results
            else:
                return False, None

        result = Promise(
            audit_results_check_func,
            "Timed out waiting for a11y audit results.",
            timeout=5,
        ).fulfill()

        # audit_results is report of accessibility violations for that session
        # Note that this ruleset doesn't have distinct error/warning levels.
        audit_results = result.get('violations')
        return audit_results

    @staticmethod
    def get_errors(audit):
        """
        Args:

            audit: results of `AxeCoreAudit.do_audit()`.

        Returns:

            A dictionary with keys "errors" and "total".
        """
        errors = {"errors": [], "total": 0}
        for session_result in audit:
            if session_result:
                errors["errors"].extend(session_result)
                for i in session_result:
                    for _node in i["nodes"]:
                        errors["total"] += 1
        return errors

    @staticmethod
    def format_errors(errors):
        """
        Args:

            errors: results of `AxeCoreAudit.get_errors()`.

        Returns: The errors as a formatted string.
        """
        lines = []
        for error_type in errors:
            lines.append("Rule ID: {}".format(error_type.get("id")))
            lines.append("Help URL: {}\n".format(error_type.get('helpUrl')))

            for node in error_type['nodes']:
                try:
                    msg = node['message']
                except KeyError:
                    try:
                        msg = node['any'][0]['message']
                    except (KeyError, IndexError):
                        msg = ''

                msg = "Message: {}".format(msg)
                html = "Html: {}".format(node.get('html').encode('utf-8'))
                target = "Target: {}".format(node.get('target'))
                fill_opts = {
                    'width': 100,
                    'initial_indent': '\t',
                    'subsequent_indent': '\t\t',
                }
                lines.append(fill(msg, **fill_opts))
                lines.append(fill(html, **fill_opts))
                lines.append(fill(target, **fill_opts))
                lines.append('\n')

        return '\n'.join(lines)

    @staticmethod
    def report_errors(audit, url):
        """
        Args:

            audit: results of `AxeCoreAudit.do_audit()`.
            url: the url of the page being audited.

        Raises: `AccessibilityError`
        """
        errors = AxeCoreAudit.get_errors(audit)
        if errors["total"] > 0:
            msg = "URL '{}' has {} errors:\n\n{}".format(
                url,
                errors["total"],
                AxeCoreAudit.format_errors(errors["errors"])
            )
            raise AccessibilityError(msg)



"""
bok-choy is a UI-level acceptance test framework
"""
import logging

# Suppress noisy loggers
NOISY_LOGGERS = ['selenium.webdriver.remote.remote_connection', 'paramiko.transport']
for log_name in NOISY_LOGGERS:
    logging.getLogger(log_name).setLevel(logging.WARNING)



"""
Use environment variables to configure Selenium remote WebDriver.
For use with SauceLabs (via SauceConnect) or local browsers.
"""
import logging
from json import dumps
import os
import socket

from needle.driver import (NeedleFirefox, NeedleChrome, NeedleIe,
                           NeedleSafari, NeedlePhantomJS)
from selenium import webdriver
from selenium.webdriver.chrome.options import Options

from bok_choy.promise import Promise

LOGGER = logging.getLogger(__name__)

REMOTE_ENV_VARS = [
    'SELENIUM_BROWSER',
    'SELENIUM_HOST',
    'SELENIUM_PORT',
]

SAUCE_ENV_VARS = REMOTE_ENV_VARS + [
    'SELENIUM_VERSION',
    'SELENIUM_PLATFORM',
    'SAUCE_USER_NAME',
    'SAUCE_API_KEY',
]


OPTIONAL_ENV_VARS = [
    'JOB_NAME',
    'BUILD_NUMBER',
]


BROWSERS = {
    'firefox': NeedleFirefox,
    'chrome': NeedleChrome,
    'internet explorer': NeedleIe,
    'safari': NeedleSafari,
    'phantomjs': NeedlePhantomJS
}


class BrowserConfigError(Exception):

    """
    Misconfiguration error in the environment variables.
    """
    pass


def save_screenshot(driver, name):
    """
    Save a screenshot of the browser.

    The location of the screenshot can be configured
    by the environment variable `SCREENSHOT_DIR`.  If not set,
    this defaults to the current working directory.

    Args:
        driver (selenium.webdriver): The Selenium-controlled browser.
        name (str): A name for the screenshot, which will be used in the output file name.

    Returns:
        None
    """
    if hasattr(driver, 'save_screenshot'):
        image_name = os.path.join(
            os.environ.get('SCREENSHOT_DIR', ''), name + '.png'
        )
        driver.save_screenshot(image_name)

    else:
        msg = (
            "Browser does not support screenshots. "
            "Could not save screenshot '{name}'"
        ).format(name=name)

        LOGGER.warning(msg)


def save_driver_logs(driver, prefix):
    """
    Save the selenium driver logs.

    The location of the driver log files can be configured
    by the environment variable `SELENIUM_DRIVER_LOG_DIR`.  If not set,
    this defaults to the current working directory.

    Args:
        driver (selenium.webdriver): The Selenium-controlled browser.
        prefix (str): A prefix which will be used in the output file names for the logs.

    Returns:
        None
    """
    log_types = ['browser', 'driver', 'client', 'server']
    for log_type in log_types:
        try:
            log = driver.get_log(log_type)
            file_name = os.path.join(
                os.environ.get('SELENIUM_DRIVER_LOG_DIR', ''), '{}_{}.log'.format(
                    prefix, log_type)
            )
            with open(file_name, 'w') as output_file:
                for line in log:
                    output_file.write("{}{}".format(dumps(line), '\n'))
        except:  # pylint: disable=bare-except
            msg = (
                "Could not save browser log of type '{log_type}'. "
                "It may be that the browser does not support it."
            ).format(log_type=log_type)

            LOGGER.warning(msg, exec_info=True)


def browser(tags=None, proxy=None):
    """
    Interpret environment variables to configure Selenium.
    Performs validation, logging, and sensible defaults.

    There are three cases:

    1. Local browsers: If the proper environment variables are not all set for the second case,
        then we use a local browser.  The environment variable `SELENIUM_BROWSER` can be set to
        specify which local browser to use, but the default is firefox.  Additionally, if a proxy
        instance is passed and the browser choice is either chrome or firefox, then the browser will
        be initialized with the proxy server set.

    2. Remote browser (not SauceLabs): Set all of the following environment variables, but not all of
        the ones needed for SauceLabs:

        * SELENIUM_BROWSER
        * SELENIUM_HOST
        * SELENIUM_PORT

    3. SauceLabs: Set all of the following environment variables:

        * SELENIUM_BROWSER
        * SELENIUM_VERSION
        * SELENIUM_PLATFORM
        * SELENIUM_HOST
        * SELENIUM_PORT
        * SAUCE_USER_NAME
        * SAUCE_API_KEY

    **NOTE:** these are the environment variables set by the SauceLabs
    Jenkins plugin.

    Optionally provide Jenkins info, used to identify jobs to Sauce:

        * JOB_NAME
        * BUILD_NUMBER

    `tags` is a list of string tags to apply to the SauceLabs
    job.  If not using SauceLabs, these will be ignored.

    Keyword Args:
        tags (list of str): Tags to apply to the SauceLabs job.  If not using SauceLabs, these will be ignored.
        proxy: A proxy instance.

    Returns:
        selenium.webdriver: The configured browser object used to drive tests

    Raises:
        BrowserConfigError: The environment variables are not correctly specified.
    """

    browser_name = os.environ.get('SELENIUM_BROWSER', 'firefox')

    def browser_check_func():
        """ Instantiate the browser and return the browser instance """
        # See https://openedx.atlassian.net/browse/TE-701
        try:
            # Get the class and kwargs required to instantiate the browser based on
            # whether we are using a local or remote one.
            if _use_remote_browser(SAUCE_ENV_VARS):
                browser_class, browser_args, browser_kwargs = _remote_browser_class(
                    SAUCE_ENV_VARS, tags)
            elif _use_remote_browser(REMOTE_ENV_VARS):
                browser_class, browser_args, browser_kwargs = _remote_browser_class(
                    REMOTE_ENV_VARS, tags)
            else:
                browser_class, browser_args, browser_kwargs = _local_browser_class(
                    browser_name)

            # If we are using a proxy, we need extra kwargs passed on intantiation.
            if proxy:
                browser_kwargs = _proxy_kwargs(browser_name, proxy, browser_kwargs)

            return True, browser_class(*browser_args, **browser_kwargs)

        except socket.error as err:
            LOGGER.debug('Failed to instantiate browser: ' + err.strerror)
            return False, None

    browser_instance = Promise(
        browser_check_func, "Browser is instantiated successfully.", timeout=30).fulfill()

    return browser_instance


def _local_browser_class(browser_name):
    """
    Returns class, kwargs, and args needed to instatiate the local browser.
    """

    # Log name of local browser
    LOGGER.info("Using local browser: %s [Default is firefox]", browser_name)

    # Get class of local browser based on name
    browser_class = BROWSERS.get(browser_name)
    if browser_class is None:
        raise BrowserConfigError(
            "Invalid browser name {name}.  Options are: {options}".format(
                name=browser_name, options=", ".join(BROWSERS.keys())))
    else:
        if browser_name == 'firefox':
            firefox_profile = webdriver.FirefoxProfile()

            # Bypasses the security prompt displayed by the browser when it attempts to
            # access a media device (e.g., a webcam)
            firefox_profile.set_preference('media.navigator.permission.disabled', True)

            browser_args = []
            browser_kwargs = {
                'firefox_profile': firefox_profile,
            }
        elif browser_name == 'chrome':
            chrome_options = Options()

            # Emulate webcam and microphone for testing purposes
            chrome_options.add_argument('--use-fake-device-for-media-stream')

            # Bypasses the security prompt displayed by the browser when it attempts to
            # access a media device (e.g., a webcam)
            chrome_options.add_argument('--use-fake-ui-for-media-stream')

            browser_args = []
            browser_kwargs = {
                'chrome_options': chrome_options,
            }
        else:
            browser_args, browser_kwargs = [], {}

        return browser_class, browser_args, browser_kwargs


def _remote_browser_class(env_vars, tags=None):
    """
    Returns class, kwargs, and args needed to instatiate the remote browser.
    """
    if tags is None:
        tags = []

    # Interpret the environment variables, raising an exception if they're
    # invalid
    envs = _required_envs(env_vars)
    envs.update(_optional_envs())

    # Turn the environment variables into a dictionary of desired capabilities
    caps = _capabilities_dict(envs, tags)

    if 'accessKey' in caps:
        LOGGER.info("Using SauceLabs: %s %s %s", caps['platform'], caps['browserName'], caps['version'])
    else:
        LOGGER.info("Using Remote Browser: %s", caps['browserName'])

    # Create and return a new Browser
    # We assume that the WebDriver end-point is running locally (e.g. using
    # SauceConnect)
    url = "http://{0}:{1}/wd/hub".format(
        envs['SELENIUM_HOST'], envs['SELENIUM_PORT'])

    browser_args = []
    browser_kwargs = {
        'command_executor': url,
        'desired_capabilities': caps,
    }

    return webdriver.Remote, browser_args, browser_kwargs


def _proxy_kwargs(browser_name, proxy, browser_kwargs={}):  # pylint: disable=dangerous-default-value
    """
    Determines the kwargs needed to set up a proxy based on the
    browser type.

    Returns: a dictionary of arguments needed to pass when
        instantiating the WebDriver instance.
    """

    proxy_dict = {
        "httpProxy": proxy.proxy,
        "proxyType": 'manual',
    }

    if browser_name == 'firefox' and 'desired_capabilities' not in browser_kwargs:
        # This one works for firefox locally
        wd_proxy = webdriver.common.proxy.Proxy(proxy_dict)
        browser_kwargs['proxy'] = wd_proxy
    else:
        # This one works with chrome, both locally and remote
        # This one works with firefox remote, but not locally
        if 'desired_capabilities' not in browser_kwargs:
            browser_kwargs['desired_capabilities'] = {}

        browser_kwargs['desired_capabilities']['proxy'] = proxy_dict

    return browser_kwargs


def _use_remote_browser(required_vars):
    """
    Returns a boolean indicating whether we should use a remote
    browser.  This means the user has made an attempt to set
    environment variables indicating they want to connect to SauceLabs
    or a remote browser.
    """
    return all([
        key in os.environ
        for key in required_vars
    ])


def _required_envs(env_vars):
    """
    Parse environment variables for required values,
    raising a `BrowserConfig` error if they are not found.

    Returns a `dict` of environment variables.
    """
    envs = {
        key: os.environ.get(key)
        for key in env_vars
    }

    # Check for missing keys
    missing = [key for key, val in envs.items() if val is None]
    if len(missing) > 0:
        msg = (
            "These environment variables must be set: " +
            ", ".join(missing)
        )
        raise BrowserConfigError(msg)

    # Check that we support this browser
    if envs['SELENIUM_BROWSER'] not in BROWSERS:
        msg = "Unsuppported browser: {0}".format(envs['SELENIUM_BROWSER'])
        raise BrowserConfigError(msg)

    return envs


def _optional_envs():
    """
    Parse environment variables for optional values,
    raising a `BrowserConfig` error if they are insufficiently specified.

    Returns a `dict` of environment variables.
    """
    envs = {
        key: os.environ.get(key)
        for key in OPTIONAL_ENV_VARS
        if key in os.environ
    }

    # If we're using Jenkins, check that we have all the required info
    if 'JOB_NAME' in envs and 'BUILD_NUMBER' not in envs:
        raise BrowserConfigError("Missing BUILD_NUMBER environment var")

    if 'BUILD_NUMBER' in envs and 'JOB_NAME' not in envs:
        raise BrowserConfigError("Missing JOB_NAME environment var")

    return envs


def _capabilities_dict(envs, tags):
    """
    Convert the dictionary of environment variables to
    a dictionary of desired capabilities to send to the
    Remote WebDriver.

    `tags` is a list of string tags to apply to the SauceLabs job.
    """
    capabilities = {
        'browserName': envs['SELENIUM_BROWSER'],
        'video-upload-on-pass': False,
        'sauce-advisor': False,
        'capture-html': True,
        'record-screenshots': True,
        'max-duration': 600,
        'public': 'public restricted',
        'tags': tags,
    }

    # Add SauceLabs specific environment vars if they are set.
    if _use_remote_browser(SAUCE_ENV_VARS):
        sauce_capabilities = {
            'platform': envs['SELENIUM_PLATFORM'],
            'version': envs['SELENIUM_VERSION'],
            'username': envs['SAUCE_USER_NAME'],
            'accessKey': envs['SAUCE_API_KEY'],
        }

        capabilities.update(sauce_capabilities)

    # Optional: Add in Jenkins-specific environment variables
    # to link Sauce output with the Jenkins job
    if 'JOB_NAME' in envs:
        jenkins_vars = {
            'build': envs['BUILD_NUMBER'],
            'name': envs['JOB_NAME'],
        }

        capabilities.update(jenkins_vars)

    return capabilities



"""
Interface for running accessibility audits on a PageObject.
"""
import json
import os
import requests

from abc import abstractmethod, abstractproperty, ABCMeta
from textwrap import dedent


class AccessibilityError(Exception):
    """
    The page violates one or more accessibility rules.
    """
    pass


class A11yAuditConfig(object):
    """
    The `A11yAuditConfig` object defines the options available in an
    accessibility ruleset.
    """
    __metaclass__ = ABCMeta

    def __init__(self, *args, **kwargs):
        super(A11yAuditConfig, self).__init__(*args, **kwargs)
        self.rules_file = None

    def set_rules_file(self, path=None):
        """
        Sets `self.rules_file` to the passed file.

        Args:
            A filepath where the JavaScript for the ruleset can be found.

        This is intended to be used in the case of using an extended or
        modified version of the ruleset. The interface and response
        format are expected to be unmodified.
        """
        if path:
            self.rules_file = os.path.abspath(path)

    @abstractmethod
    def set_rules(self, rules):
        """
        Overrides the default rules to be run.

        Raises:

            `NotImplementedError` if this isn't overwritten in the ruleset
                specific implementation.
        """
        raise NotImplementedError(
            "The ability to specify rules has not been implemented."
        )

    @abstractmethod
    def set_scope(self, include=None, exclude=None):
        """
        Overrides the default scope (part of the DOM) to inspect.

        Raises:

            `NotImplementedError` if this isn't overwritten in the ruleset
                specific implementation.
        """
        raise NotImplementedError(
            "The ability to specify scope has not been implemented."
        )


class A11yAudit(object):
    """
    Allows auditing of a page for accessibility issues.

    The ruleset to use can be specified by the environment variable
    `BOKCHOY_A11Y_RULESET`. Currently, there are two ruleset implemented:

    `axe_core`:

        * Ruleset class: AxeCoreAudit
        * Ruleset config: AxeCoreAuditConfig
        * This is default ruleset.

    `google_axs`:

        * Ruleset class: AxsAudit
        * Ruleset config: AxsAuditConfig
    """
    __metaclass__ = ABCMeta

    def __init__(self, browser, url, config=None, *args, **kwargs):
        """
        Sets ruleset to be used.

        Args:
            browser: A phantomjs browser instance
            url: URL of the page to test
            config: (optional) A11yAuditConfig or subclass of A11yAuditConfig
        """
        super(A11yAudit, self).__init__(*args, **kwargs)
        self.url = url
        self.browser = browser
        self.config = config or self.default_config

    def _phantomjs_setup(self):
        """
        Verifies that phantomjs is being used and returns the ghostdriver URL
        and session information.

        Returns: (ghostdriver_url, sessions) where sessions is a list of
            session ids.

        Raises: `NotImplementedError` if not using phantomjs.
        """
        if self.browser.name != 'phantomjs':
            msg = (
                'Accessibility auditing is only supported with PhantomJS as'
                ' the browser.'
            )
            raise NotImplementedError(msg)

        # The ghostdriver URL will be something like this:
        # 'http://localhost:33225/wd/hub'
        ghostdriver_url = self.browser.service.service_url

        # Get the session_id from ghostdriver so that we can inject JS into
        # the page.
        resp = requests.get('{}/sessions'.format(ghostdriver_url))
        sessions = resp.json()

        return ghostdriver_url, sessions

    def _verify_rules_file_exists(self):
        """
        Checks that the rules file for the enabled ruleset exists.

        Raises: `RuntimeError` if the file isn't found.
        """
        if not os.path.isfile(self.config.rules_file):
            msg = 'Could not find the accessibility tools JS file: {}'.format(
                self.config.rules_file)
            raise RuntimeError(msg)

    def do_audit(self):
        """
        Audit the page for accessibility problems using the enabled ruleset.

        Since this needs to inject JavaScript into the browser page, the only
        known way to do this is to use PhantomJS as your browser.

        Raises:

            * NotImplementedError if you are not using PhantomJS
            * RuntimeError if there was a problem with the injected JS or
                getting the report

        Returns:
            A list (one for each browser session) of results returned from
            the audit. See documentation of `_check_rules` in the enabled
            ruleset for the format of each result.
            `None` if no results are returned.
        """
        ghostdriver_url, sessions = self._phantomjs_setup()
        self._verify_rules_file_exists()

        # report is the list that is returned, with one item for each
        # browser session.
        report = []
        for session in sessions.get('value'):
            session_id = session.get('id')

            # First make sure you can successfully inject the JS on the page
            script = dedent("""
                return this.injectJs("{file}");
            """.format(file=self.config.rules_file))

            payload = {"script": script, "args": []}
            resp = requests.post('{}/session/{}/phantom/execute'.format(
                ghostdriver_url, session_id), data=json.dumps(payload))

            result = resp.json().get('value')

            if result is False:
                msg = '{msg} \nScript:{script} \nResponse:{response}'.format(
                    msg=(
                        'Failure injecting the Accessibility Audit JS '
                        'on the page.'
                    ),
                    script=script,
                    response=resp.text)
                raise RuntimeError(msg)

            audit_results = self._check_rules(
                ghostdriver_url, session_id, self.config)

            if audit_results:
                report.append(audit_results)

        return report or None

    def check_for_accessibility_errors(self):
        """
        Run an accessibility audit, parse the results, and raise a single
        exception if there are violations.

        Note that an exception is only raised on errors, not on warnings.

        Returns:

            None

        Raises:

            AccessibilityError
        """
        audit_results = self.do_audit()
        if audit_results:
            self.report_errors(audit_results, self.url)

    @abstractproperty
    def default_config(self):
        """
        Return an instance of a subclass of A11yAuditConfig.
        """
        raise NotImplementedError("default_config has not been implemented")

    @staticmethod
    @abstractmethod
    def _check_rules(ghostdriver_url, session_id, config):
        """
        Run an accessibility audit on the page using the implemented ruleset.

        Args:

            ghostdriver_url: url of ghostdriver.
            session_id: a session id to test.
            config: an AxsAuditConfig instance.

        Returns:

            A list of violations.

        Raises:

            `NotImplementedError` if this isn't overwritten in the ruleset
                specific implementation.
        """
        raise NotImplementedError("_check_rules has not been implemented")

    @staticmethod
    @abstractmethod
    def report_errors(audit, url):
        """
        Args:

            audit: results of an accessibility audit.
            url: the url of the page being audited.

        Raises:

            `AccessibilityError` if errors are found in the audit.
            `NotImplementedError` if this isn't overwritten in the ruleset
                specific implementation.
        """
        raise NotImplementedError("report_errors has not been implemented")



"""
Utility methods for using browsermobproxy.
"""
import os
import signal
import psutil
from browsermobproxy import Server
from .promise import Promise


def bmp_proxy():
    """
    Creates a proxy and a server instance with browsermobproxy.
    Reference: http://browsermob-proxy-py.readthedocs.org/en/latest/index.html

    Returns:
        (proxy, server)
    """
    def create_proxy():
        """
        Try to create a proxy.
        """
        try:
            proxy = server.create_proxy()
        except:  # pylint: disable=bare-except
            return False, None
        return True, proxy

    port = int(os.environ.get('BROWSERMOB_PROXY_PORT', 8080))
    server = Server('browsermob-proxy', options={'port': port})

    try:
        # If anything in this block raises an exception, make sure we kill
        # the server process before exiting.
        server.start()

        # Using the promise module to wait for the server to be responsive.
        # The server.create_proxy function sometimes raises connection
        # refused errors if the server isn't ready yet.
        proxy = Promise(
            create_proxy, 'browsermobproxy is responsive', timeout=10
        ).fulfill()

        proxy_host = os.environ.get('BROWSERMOB_PROXY_HOST', '127.0.0.1')
        proxy.remap_hosts('localhost', proxy_host)
    except:  # pylint: disable=bare-except
        # Make sure that the server process is stopped.
        stop_server(server)
        raise

    return proxy, server


def kill_process(proc):
    """
    Kill the process `proc` created with `subprocess`.

    Args:
        process
    Returns:
        None
    """
    p1_group = psutil.Process(proc.pid)

    child_pids = p1_group.children(recursive=True)

    for child_pid in child_pids:
        os.kill(child_pid.pid, signal.SIGKILL)


def stop_server(server):
    """
    Stops the browsermobproxy server process and any child_pid
    processes.

    Args:
        server
    Returns:
        None
    """
    # Server.stop does not kill the child processes, but it needs to.
    # Do not remove this or the processes will not get killed.
    kill_process(server.process)
    server.stop()



"""
Interface for using the google accessibility ruleset.
See: https://github.com/GoogleChrome/accessibility-developer-tools
"""
import json
import logging
import os
import requests

from collections import namedtuple
from textwrap import dedent

from .a11y_audit import A11yAudit, A11yAuditConfig, AccessibilityError

log = logging.getLogger(__name__)
AuditResults = namedtuple('AuditResults', 'errors, warnings')
CUR_DIR = os.path.dirname(os.path.abspath(__file__))


class AxsAuditConfig(A11yAuditConfig):
    """
    The `AxsAuditConfig` object defines the options available when
    running an `AxsAudit`.
    """
    def __init__(self, *args, **kwargs):
        super(AxsAuditConfig, self).__init__(*args, **kwargs)
        self.rules_file = None
        self.scope, self.rules_to_run, self.rules_to_ignore = None, None, None
        self.rules_file = os.path.join(
            os.path.split(CUR_DIR)[0],
            'vendor/google/axs_testing.js'
        )
        self.set_rules({})
        self.set_scope()

    def set_rules(self, rules):
        """
        Sets the rules to be run or ignored for the audit.

        Args:

            rules: a dictionary of the format `{"ignore": [], "apply": []}`.

        See https://github.com/GoogleChrome/accessibility-developer-tools/tree/master/src/audits

        Passing `{"apply": []}` or `{}` means to check for all available rules.

        Passing `{"apply": None}` means that no audit should be done for this page.

        Passing `{"ignore": []}` means to run all otherwise enabled rules.
        Any rules in the "ignore" list will be ignored even if they were also
        specified in the "apply".

        Examples:

            To check only `badAriaAttributeValue`::

                page.a11y_audit.config.set_rules({
                    "apply": ['badAriaAttributeValue']
                })

            To check all rules except `badAriaAttributeValue`::

                page.a11y_audit.config.set_rules(
                    "ignore": ['badAriaAttributeValue'],
                )
        """
        self.rules_to_ignore = rules.get("ignore", [])
        self.rules_to_run = rules.get("apply", [])

    def set_scope(self, include=None, exclude=None):
        """
        Sets `scope`, the "start point" for the audit.

        Args:

            include: A list of css selectors specifying the elements that
                contain the portion of the page that should be audited.
                Defaults to auditing the entire document.
            exclude: This arg is not implemented in this ruleset.

        Examples:

            To check only the `div` with id `foo`::

                page.a11y_audit.config.set_scope(["div#foo"])

            To reset the scope to check the whole document::

                page.a11y_audit.config.set_scope()
        """
        if include:
            self.scope = "document.querySelector(\"{}\")".format(
                ', '.join(include)
            )
        else:
            self.scope = "null"

        if exclude is not None:
            raise NotImplementedError(
                "The argument `exclude` has not been implemented in "
                "AxsAuditConfig.set_scope method."
            )


class AxsAudit(A11yAudit):
    """
    Use Google's Accessibility Developer Tools to audit a
    page for accessibility problems.

    See https://github.com/GoogleChrome/accessibility-developer-tools

    Since this needs to inject JavaScript into the browser page, the only
    known way to do this is to use PhantomJS as your browser.
    """

    @property
    def default_config(self):
        """
        Returns an instance of a subclass of AxsAuditConfig.
        """
        return AxsAuditConfig()

    @staticmethod
    def _check_rules(ghostdriver_url, session_id, config):
        """
        Check the page for violations of the configured rules. By default,
        all rules in the ruleset will be checked.

        Args:
            ghostdriver_url: url of ghostdriver.
            session_id: a session id to test.
            config: an AxsAuditConfig instance.

        Returns:
            A namedtuple with 'errors' and 'warnings' fields whose values are
            the errors and warnings returned from the audit.

            None if config has rules_to_run set to None.

        __Caution__: You probably don't really want to call this method
        directly! It will be used by `A11yAudit.do_audit` if using this ruleset.
        """
        if config.rules_to_run is None:
            msg = 'No accessibility rules were specified to check.'
            log.warning(msg)
            return None

        # This line will only be included in the script if rules to check on
        # this page are specified, as the default behavior of the js is to
        # run all rules.
        rules = config.rules_to_run
        if len(rules) > 0:
            rules_config = "auditConfig.auditRulesToRun = {rules};".format(
                rules=rules)
        else:
            rules_config = ""

        ignored_rules = config.rules_to_ignore
        if ignored_rules:
            rules_config += (
                "\nauditConfig.auditRulesToIgnore = {rules};".format(
                    rules=ignored_rules
                )
            )

        script = dedent("""
            return this.evaluate(function() {{
              var auditConfig = new axs.AuditConfiguration();
              {rules_config}
              auditConfig.scope = {scope};
              var run_results = axs.Audit.run(auditConfig);
              var audit_results = axs.Audit.auditResults(run_results)
              return audit_results;
            }});
        """.format(rules_config=rules_config, scope=config.scope))

        payload = {"script": script, "args": []}
        resp = requests.post('{}/session/{}/phantom/execute'.format(
            ghostdriver_url, session_id), data=json.dumps(payload))

        result = resp.json().get('value')
        if result is None:
            msg = '{} {} \nScript:{} \nResponse:{}'.format(
                'No results were returned by the audit report.',
                (
                    'Perhaps there was a problem with the rules or scope '
                    'defined for this page.'
                ),
                script,
                resp.text)
            raise RuntimeError(msg)

        # audit_results is report of accessibility errors for that session
        audit_results = AuditResults(
            errors=result.get('errors_'),
            warnings=result.get('warnings_')
        )
        return audit_results

    @staticmethod
    def get_errors(audit):
        """
        Args:

            audit: results of `AxsAudit.do_audit()`.

        Returns: a list of errors.
        """
        errors = []
        for session_result in audit:
            if session_result:
                if session_result.errors:
                    errors.extend(session_result.errors)

        return errors

    @staticmethod
    def report_errors(audit, url):
        """
        Args:

            audit: results of `AxsAudit.do_audit()`.
            url: the url of the page being audited.

        Raises: `AccessibilityError`
        """
        errors = AxsAudit.get_errors(audit)
        if errors:
            msg = "URL '{}' has {} errors:\n{}".format(
                url,
                len(errors),
                (', ').join(errors)
            )
            raise AccessibilityError(msg)
